import glob, cv2, torch
import numpy as np

from network.SiamFCOSP import SiamFCOSP_eval
from run_Siamfcos import SiamFCOS_init, SiamFCOS_track
from utils import get_axis_aligned_bbox, cxy_wh_2_rect

# load net
net = SiamFCOSP_eval()
net.load_state_dict(torch.load('SiamFCOSPlus_net_epoch15')['network'])
net.eval()

image_files = sorted(glob.glob('./bag/*.jpg'))
init_rbox = [334.02,128.36,438.19,188.78,396.39,260.83,292.23,200.41]

[cx, cy, w, h] = get_axis_aligned_bbox(init_rbox)

# tracker init
target_pos, target_sz = np.array([cx, cy]), np.array([w, h])
im = cv2.imread(image_files[0])  # HxWxC
state = SiamFCOS_init(im, target_pos, target_sz, net)

# tracking and visualization
toc = 0
for f, image_file in enumerate(image_files):
    im = cv2.imread(image_file)
    tic = cv2.getTickCount()
    state = SiamFCOS_track(state, im)  # track
    toc += cv2.getTickCount()-tic
    res = cxy_wh_2_rect(state['target_pos'], state['target_sz'])
    res = [int(l) for l in res]
    cv2.rectangle(im, (res[0], res[1]), (res[0] + res[2], res[1] + res[3]), (0, 255, 255), 3)
    cv2.imshow('SiamFCOS', im)
    cv2.waitKey(0)

print('Tracking Speed {:.1f}fps'.format((len(image_files)-1)/(toc/cv2.getTickFrequency())))
