import os

# import ipdb
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.module import Module


class RPN(nn.Module):
    def __init__(self):
        super(RPN, self).__init__()
        self.adj_1 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))

        self.adj_2 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))
        self.adj_3 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))
        self.adj_4 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))

        #
        self.fusion_module_1 = nn.Sequential(nn.Conv2d(256, 256, kernel_size=1, padding=0, stride=1),
                                             nn.BatchNorm2d(256),
                                             nn.ReLU(inplace=True))

        self.fusion_module_2 = nn.Sequential(nn.Conv2d(256, 256, kernel_size=1, padding=0, stride=1),
                                             nn.BatchNorm2d(256),
                                             nn.ReLU(inplace=True))

        self.Box_Head = nn.Sequential(nn.Conv2d(256, 4 * 5, kernel_size=1, padding=0, stride=1))
        # self.Box_Head = nn.Sequential(nn.Conv2d(256, 4, kernel_size=1, padding=0, stride=1))
        self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2 * 5, kernel_size=1, padding=0, stride=1))
        # self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2, kernel_size=1, padding=0, stride=1))
        # self.centerness_Head = nn.Sequential(nn.Conv2d(256, 1, kernel_size=1, padding=0, stride=1))

    def forward(self, examplar_feature_map, search_region_feature_map, BATCH_SIZE):
        # crop
        cropped_examplar_feature_map = F.pad(examplar_feature_map, (-4, -4, -4, -4))

        # TODO
        adj_1_output = self.adj_1(cropped_examplar_feature_map)
        adj_1_output = adj_1_output.reshape(-1, 7, 7)
        adj_1_output = adj_1_output.unsqueeze(0).permute(1, 0, 2, 3)

        adj_2_output = self.adj_2(search_region_feature_map)
        adj_2_output = adj_2_output.reshape(-1, 31, 31)
        adj_2_output = adj_2_output.unsqueeze(0)

        # TODO
        adj_3_output = self.adj_3(cropped_examplar_feature_map)
        adj_3_output = adj_3_output.reshape(-1, 7, 7)
        adj_3_output = adj_3_output.unsqueeze(0).permute(1, 0, 2, 3)

        adj_4_output = self.adj_4(search_region_feature_map)
        adj_4_output = adj_4_output.reshape(-1, 31, 31)
        adj_4_output = adj_4_output.unsqueeze(0)

        depthwise_cross_reg = F.conv2d(
            adj_2_output, adj_1_output, bias=None, stride=1, padding=0, groups=adj_1_output.size()[0]).squeeze()

        depthwise_cross_cls = F.conv2d(
            adj_4_output, adj_3_output, bias=None, stride=1, padding=0, groups=adj_3_output.size()[0]).squeeze()

        # ipdb.set_trace()
        depthwise_cross_reg = depthwise_cross_reg.reshape(-1, 256, 25, 25)
        depthwise_cross_cls = depthwise_cross_cls.reshape(-1, 256, 25, 25)

        depthwise_cross_reg = self.fusion_module_1(depthwise_cross_reg)
        depthwise_cross_cls = self.fusion_module_2(depthwise_cross_cls)

        bbox_regression_prediction = self.Box_Head(depthwise_cross_reg)
        cls_prediction = self.Cls_Head(depthwise_cross_cls)
        # centerness_prediction = self.centerness_Head(depthwise_cross_cls)
        return cls_prediction, bbox_regression_prediction


class RPN_FCOS(nn.Module):
    def __init__(self):
        super(RPN_FCOS, self).__init__()
        self.adj_1 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))

        self.adj_2 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))
        self.adj_3 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))
        self.adj_4 = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, padding=1, stride=1), nn.BatchNorm2d(256))

        #
        self.fusion_module_1 = nn.Sequential(nn.Conv2d(256, 256, kernel_size=1, padding=0, stride=1),
                                             nn.BatchNorm2d(256),
                                             nn.ReLU(inplace=True))

        self.fusion_module_2 = nn.Sequential(nn.Conv2d(256, 256, kernel_size=1, padding=0, stride=1),
                                             nn.BatchNorm2d(256),
                                             nn.ReLU(inplace=True))

        # self.Box_Head = nn.Sequential(nn.Conv2d(256, 4 * 5, kernel_size=1, padding=0, stride=1))
        self.Box_Head = nn.Sequential(nn.Conv2d(256, 4, kernel_size=1, padding=0, stride=1))
        # self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2 * 5, kernel_size=1, padding=0, stride=1))
        self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2, kernel_size=1, padding=0, stride=1))
        self.centerness_Head = nn.Sequential(nn.Conv2d(256, 1, kernel_size=1, padding=0, stride=1))

        # self.weighted_sum_layer_alpha = nn.Conv2d(6, 2, kernel_size=1, padding=0, groups=2)
        # self.weighted_sum_layer_beta = nn.Conv2d(
        #     60, 20, kernel_size=1, padding=0, groups=20)
        # self.weighted_sum_layer_beta = nn.Conv2d(12, 4, kernel_size=1, padding=0, groups=4)
        # self.weighted_sum_layer_omega = nn.Conv2d(3, 1, kernel_size=1, padding=0, groups=1)
        # nn.init.normal_(self.weighted_sum_layer_alpha.weight, std=0.01)
        # nn.init.constant_(self.weighted_sum_layer_alpha.bias, 0)
        # nn.init.normal_(self.weighted_sum_layer_beta.weight, std=0.01)
        # nn.init.constant_(self.weighted_sum_layer_beta.bias, 0)
        # nn.init.normal_(self.weighted_sum_layer_omega.weight, std=0.01)
        # nn.init.constant_(self.weighted_sum_layer_omega.bias, 0)

    def forward(self, examplar_feature_map, search_region_feature_map, BATCH_SIZE):
        # crop
        cropped_examplar_feature_map = F.pad(examplar_feature_map, (-4, -4, -4, -4))

        # TODO
        adj_1_output = self.adj_1(cropped_examplar_feature_map)
        adj_1_output = adj_1_output.reshape(-1, 7, 7)
        adj_1_output = adj_1_output.unsqueeze(0).permute(1, 0, 2, 3)

        adj_2_output = self.adj_2(search_region_feature_map)
        adj_2_output = adj_2_output.reshape(-1, 31, 31)
        adj_2_output = adj_2_output.unsqueeze(0)

        # TODO
        adj_3_output = self.adj_3(cropped_examplar_feature_map)
        adj_3_output = adj_3_output.reshape(-1, 7, 7)
        adj_3_output = adj_3_output.unsqueeze(0).permute(1, 0, 2, 3)

        adj_4_output = self.adj_4(search_region_feature_map)
        adj_4_output = adj_4_output.reshape(-1, 31, 31)
        adj_4_output = adj_4_output.unsqueeze(0)

        depthwise_cross_reg = F.conv2d(
            adj_2_output, adj_1_output, bias=None, stride=1, padding=0, groups=adj_1_output.size()[0]).squeeze()

        depthwise_cross_cls = F.conv2d(
            adj_4_output, adj_3_output, bias=None, stride=1, padding=0, groups=adj_3_output.size()[0]).squeeze()

        # ipdb.set_trace()
        depthwise_cross_reg = depthwise_cross_reg.reshape(-1, 256, 25, 25)
        depthwise_cross_cls = depthwise_cross_cls.reshape(-1, 256, 25, 25)

        depthwise_cross_reg = self.fusion_module_1(depthwise_cross_reg)
        depthwise_cross_cls = self.fusion_module_2(depthwise_cross_cls)

        bbox_regression_prediction = self.Box_Head(depthwise_cross_reg)
        cls_prediction = self.Cls_Head(depthwise_cross_cls)
        centerness_prediction = self.centerness_Head(depthwise_cross_cls)
        return cls_prediction, bbox_regression_prediction, centerness_prediction


class RPN_Raw(nn.Module):
    def __init__(self):
        super(RPN_Raw, self).__init__()

        self.fusion_module_1 = nn.Sequential(nn.Conv2d(256, 256, kernel_size=7, padding=0, stride=1),
                                             nn.BatchNorm2d(256),
                                             nn.ReLU(inplace=True))

        # self.Box_Head = nn.Sequential(nn.Conv2d(256, 4 * 5, kernel_size=1, padding=0, stride=1))
        self.Box_Head = nn.Sequential(nn.Conv2d(256, 4, kernel_size=1, padding=0, stride=1))
        # self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2 * 5, kernel_size=1, padding=0, stride=1))
        self.Cls_Head = nn.Sequential(nn.Conv2d(256, 2, kernel_size=1, padding=0, stride=1))
        self.centerness_Head = nn.Sequential(nn.Conv2d(256, 1, kernel_size=1, padding=0, stride=1))

    def forward(self, search_region_feature_map):
        # crop
        depthwise_cross_reg = self.fusion_module_1(search_region_feature_map)
        bbox_regression_prediction = self.Box_Head(depthwise_cross_reg)
        cls_prediction = self.Cls_Head(depthwise_cross_reg)
        centerness_prediction = self.centerness_Head(depthwise_cross_reg)
        return cls_prediction, bbox_regression_prediction, centerness_prediction

