from network.customized_resnet import *
from network.RPN import *
import torch.nn as nn
import matplotlib.pyplot

class SiamFCOSP(nn.Module):
    def __init__(self):
        super(SiamFCOSP, self).__init__()
        # self.examplar_branch = resnet50()
        # self.search_region_branch = resnet50()

        self.feature_branch = resnet50(pretrained=False)
        self.examplar_branch = self.feature_branch
        self.search_region_branch = self.feature_branch
        self.conv3_3_RPN = RPN_FCOS()
        self.conv4_6_RPN = RPN_FCOS()
        self.conv5_3_RPN = RPN_FCOS()

        self.conv3_RPN_RAW = RPN_Raw()
        self.conv4_RPN_RAW = RPN_Raw()
        self.conv5_RPN_RAW = RPN_Raw()
        # self.weighted_sum_layer_alpha = nn.Conv2d(
        #     30, 10, kernel_size=1, padding=0, groups=10)
        self.weighted_sum_layer_alpha = nn.Conv2d(12, 2, kernel_size=1, padding=0, groups=2)
        # self.weighted_sum_layer_beta = nn.Conv2d(
        #     60, 20, kernel_size=1, padding=0, groups=20)
        self.weighted_sum_layer_beta = nn.Conv2d(24, 4, kernel_size=1, padding=0, groups=4)
        self.weighted_sum_layer_omega = nn.Conv2d(6, 1, kernel_size=1, padding=0, groups=1)
        nn.init.normal_(self.weighted_sum_layer_alpha.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_alpha.bias, 0)
        nn.init.normal_(self.weighted_sum_layer_beta.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_beta.bias, 0)
        nn.init.normal_(self.weighted_sum_layer_omega.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_omega.bias, 0)
        self.scale = Scale()

    def forward(self, examplar, search_region):
        _, examplar_conv_3_output, examplar_conv_4_output, examplar_conv_5_output = self.examplar_branch(examplar)

        _, search_region_conv_3_output, search_region_conv_4_output, search_region_conv_5_output = self.search_region_branch(search_region)

        conv3_3_cls_prediction, conv3_3_bbox_regression_prediction, conv3_3_centerness_prediction = self.conv3_3_RPN(
            examplar_conv_3_output, search_region_conv_3_output, examplar.size()[0])
        conv4_6_cls_prediction, conv4_6_bbox_regression_prediction, conv4_6_centerness_prediction = self.conv4_6_RPN(
            examplar_conv_4_output, search_region_conv_4_output, examplar.size()[0])
        conv5_3_cls_prediction, conv5_3_bbox_regression_prediction, conv5_3_centerness_prediction = self.conv5_3_RPN(
            examplar_conv_5_output, search_region_conv_5_output, examplar.size()[0])

        conv3_raw_cls_prediction, conv3_raw_bbox_regression_prediction, conv3_raw_centerness_prediction = self.conv3_RPN_RAW(
            search_region_conv_3_output)
        conv4_raw_cls_prediction, conv4_raw_bbox_regression_prediction, conv4_raw_centerness_prediction = self.conv4_RPN_RAW(
            search_region_conv_4_output)
        conv5_raw_cls_prediction, conv5_raw_bbox_regression_prediction, conv5_raw_centerness_prediction = self.conv5_RPN_RAW(
            search_region_conv_5_output)
        # ipdb.set_trace()
        stacked_cls_prediction = torch.cat((conv3_3_cls_prediction, conv4_6_cls_prediction, conv5_3_cls_prediction,
                                            conv3_raw_cls_prediction, conv4_raw_cls_prediction, conv5_raw_cls_prediction), 2).reshape(
            examplar.size()[0], 2, -1, 25, 25).reshape(examplar.size()[0], -1, 25, 25)

        stacked_regression_prediction = torch.cat((conv3_3_bbox_regression_prediction, conv4_6_bbox_regression_prediction, conv5_3_bbox_regression_prediction,
                                                   conv3_raw_bbox_regression_prediction, conv4_raw_bbox_regression_prediction, conv5_raw_bbox_regression_prediction), 2).reshape(
            examplar.size()[0], 4, -1, 25, 25).reshape(examplar.size()[0], -1, 25, 25)

        stacked_centerness_prediction = torch.cat((conv3_3_centerness_prediction, conv4_6_centerness_prediction, conv5_3_centerness_prediction,
                                                   conv3_raw_centerness_prediction, conv4_raw_centerness_prediction, conv5_raw_centerness_prediction),).reshape(
            examplar.size()[0], 1, -1, 25, 25).reshape(examplar.size()[0], -1, 25, 25)

        # ipdb.set_trace()
        fused_cls_prediction = self.weighted_sum_layer_alpha(stacked_cls_prediction)
        fused_regression_prediction = self.weighted_sum_layer_beta(stacked_regression_prediction)
        fused_centerness_prediction = self.weighted_sum_layer_omega(stacked_centerness_prediction)

        # cls_prediction = F.softmax(fused_cls_prediction, dim=1)
        centerness_prediction = fused_centerness_prediction
        bbox_reg = torch.exp(self.scale(fused_regression_prediction))
        return fused_cls_prediction, bbox_reg, centerness_prediction
        # return cls_prediction, bbox_reg, centerness_prediction


class Scale(nn.Module):
    def __init__(self, init_value=1.0):
        super(Scale, self).__init__()
        self.scale = nn.Parameter(torch.FloatTensor([init_value]))

    def forward(self, input):
        return input * self.scale


class SiamFCOSP_eval(nn.Module):
    def __init__(self):
        super(SiamFCOSP_eval, self).__init__()
        # self.examplar_branch = resnet50()
        # self.search_region_branch = resnet50()
        self.feature_branch = resnet50()
        self.examplar_branch = self.feature_branch
        self.search_region_branch = self.feature_branch
        self.conv3_3_RPN = RPN_FCOS()
        self.conv4_6_RPN = RPN_FCOS()
        self.conv5_3_RPN = RPN_FCOS()

        self.conv3_RPN_RAW = RPN_Raw()
        self.conv4_RPN_RAW = RPN_Raw()
        self.conv5_RPN_RAW = RPN_Raw()
        # self.weighted_sum_layer_alpha = nn.Conv2d(
        #     30, 10, kernel_size=1, padding=0, groups=10)
        self.weighted_sum_layer_alpha = nn.Conv2d(12, 2, kernel_size=1, padding=0, groups=2)
        # self.weighted_sum_layer_beta = nn.Conv2d(
        #     60, 20, kernel_size=1, padding=0, groups=20)
        self.weighted_sum_layer_beta = nn.Conv2d(24, 4, kernel_size=1, padding=0, groups=4)
        self.weighted_sum_layer_omega = nn.Conv2d(6, 1, kernel_size=1, padding=0, groups=1)
        nn.init.normal_(self.weighted_sum_layer_alpha.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_alpha.bias, 0)
        nn.init.normal_(self.weighted_sum_layer_beta.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_beta.bias, 0)
        nn.init.normal_(self.weighted_sum_layer_omega.weight, std=0.01)
        nn.init.constant_(self.weighted_sum_layer_omega.bias, 0)
        self.scale = Scale()

    def temple(self, examplar):
        self.examplar_size = examplar.size()
        _, self.examplar_conv_3_output, self.examplar_conv_4_output, self.examplar_conv_5_output = self.examplar_branch(examplar)

    def forward(self, search_region):
        # _, examplar_conv_3_output, examplar_conv_4_output, examplar_conv_5_output = self.examplar_branch(examplar)

        _, search_region_conv_3_output, search_region_conv_4_output, search_region_conv_5_output = self.search_region_branch(search_region)

        conv3_3_cls_prediction, conv3_3_bbox_regression_prediction, conv3_3_centerness_prediction = self.conv3_3_RPN(
            self.examplar_conv_3_output, search_region_conv_3_output, self.examplar_size[0])
        conv4_6_cls_prediction, conv4_6_bbox_regression_prediction, conv4_6_centerness_prediction = self.conv4_6_RPN(
            self.examplar_conv_4_output, search_region_conv_4_output, self.examplar_size[0])
        conv5_3_cls_prediction, conv5_3_bbox_regression_prediction, conv5_3_centerness_prediction = self.conv5_3_RPN(
            self.examplar_conv_5_output, search_region_conv_5_output, self.examplar_size[0])

        conv3_raw_cls_prediction, conv3_raw_bbox_regression_prediction, conv3_raw_centerness_prediction = self.conv3_RPN_RAW(
            search_region_conv_3_output)
        conv4_raw_cls_prediction, conv4_raw_bbox_regression_prediction, conv4_raw_centerness_prediction = self.conv4_RPN_RAW(
            search_region_conv_4_output)
        conv5_raw_cls_prediction, conv5_raw_bbox_regression_prediction, conv5_raw_centerness_prediction = self.conv5_RPN_RAW(
            search_region_conv_5_output)
        # ipdb.set_trace()
        stacked_cls_prediction = torch.cat((conv3_3_cls_prediction, conv4_6_cls_prediction, conv5_3_cls_prediction,
                                            conv3_raw_cls_prediction, conv4_raw_cls_prediction,
                                            conv5_raw_cls_prediction), 2).reshape(
            self.examplar_size[0], 2, -1, 25, 25).reshape(self.examplar_size[0], -1, 25, 25)

        stacked_regression_prediction = torch.cat(
            (conv3_3_bbox_regression_prediction, conv4_6_bbox_regression_prediction, conv5_3_bbox_regression_prediction,
             conv3_raw_bbox_regression_prediction, conv4_raw_bbox_regression_prediction,
             conv5_raw_bbox_regression_prediction), 2).reshape(
            self.examplar_size[0], 4, -1, 25, 25).reshape(self.examplar_size[0], -1, 25, 25)

        stacked_centerness_prediction = torch.cat(
            (conv3_3_centerness_prediction, conv4_6_centerness_prediction, conv5_3_centerness_prediction,
             conv3_raw_centerness_prediction, conv4_raw_centerness_prediction,
             conv5_raw_centerness_prediction), ).reshape(
            self.examplar_size[0], 1, -1, 25, 25).reshape(self.examplar_size[0], -1, 25, 25)

        # ipdb.set_trace()
        fused_cls_prediction = self.weighted_sum_layer_alpha(stacked_cls_prediction)
        fused_regression_prediction = self.weighted_sum_layer_beta(stacked_regression_prediction)
        fused_centerness_prediction = self.weighted_sum_layer_omega(stacked_centerness_prediction)

        centerness_prediction = fused_centerness_prediction.sigmoid()
        bbox_reg = torch.exp(self.scale(fused_regression_prediction))
        return fused_cls_prediction, bbox_reg, centerness_prediction


